"""Unit tests for optimizers"""
import numpy as np

from pysensors.optimizers import CCQR
from pysensors.optimizers import QR


def test_num_sensors(data_vandermonde):
    x = data_vandermonde
    max_sensors = x.shape[1]

    qr = QR()
    sensors = qr.fit(x.T).get_sensors()
    assert len(sensors) == max_sensors


def test_ccqr_qr_equivalence(data_vandermonde):
    x = data_vandermonde

    qr_sensors = QR().fit(x.T).get_sensors()
    # If no costs are passed, all zeros are used
    ccqr_sensors = CCQR().fit(x.T).get_sensors()

    np.testing.assert_array_equal(qr_sensors, ccqr_sensors)


def test_ccqr_sensor_placement(data_random):
    x = data_random

    forbidden_sensors = np.arange(0, x.shape[1], 3)
    costs = np.zeros(x.shape[1])
    costs[forbidden_sensors] = 100
    # Get ranked sensors
    sensors = CCQR(sensor_costs=costs).fit(x.T).get_sensors()

    # Forbidden sensors should not be included
    chosen_sensors = set(sensors[: (x.shape[1] - len(forbidden_sensors))])
    assert chosen_sensors.isdisjoint(set(forbidden_sensors))


def test_ccqr_negative_costs(data_vandermonde):
    x = data_vandermonde

    desirable_sensors = np.array([20, 55, 99, 100, 150])
    costs = np.zeros(x.shape[1])
    costs[desirable_sensors] = -100

    sensors = CCQR(sensor_costs=costs).fit(x.T).get_sensors()

    chosen_sensors = set(sensors[: min(x.shape)])
    assert all(s in chosen_sensors for s in set(desirable_sensors))



import numpy as np

from ._qr import QR


class CCQR(QR):
    """
    Greedy cost-constrained QR optimizer for sensor selection.
    Ranks sensors in descending order of "importance" based on
    reconstruction performance and sensor cost. Uses a cost-sensitive
    version of the QR algorithm.

    This method is based on the following work:

        Clark, Emily, et al.
        "Greedy sensor placement with cost constraints."
        IEEE Sensors Journal 19.7 (2018): 2642-2656.
    """

    def __init__(self, sensor_costs=None):
        """
        Greedy cost-constrained QR optimizer for sensor selection.
        This algorithm augments the pivot selection criteria used in the
        QR algorithm (with householder reflectors) to take into account
        costs associated with each sensors. It is similar to the
        :class:`pysensors.optimizers.QR` algorithm in that it returns an array
        of sensor locations ranked by importance, but with a definition of
        importance that takes sensor costs into account.

        Parameters
        ----------
        sensor_costs: np.ndarray, shape [n_features,], optional (default None)
            Costs (weights) associated with each sensor.
            Positive values will encourage sensors to be avoided and
            negative values will cause them to be preferred.
            If None, costs will all be set to zero.

        Attributes
        ----------
        pivots_ : np.ndarray, shape [n_features]
            Ranked list of sensor locations.
        """
        super(CCQR, self).__init__()
        if sensor_costs is not None and np.ndim(sensor_costs) != 1:
            raise ValueError(
                "sensor_costs must be a 1D array, "
                f"but a {np.ndim(sensor_costs)}D array was given"
            )
        self.sensor_costs = sensor_costs

    def fit(
        self,
        basis_matrix,
    ):
        """
        Parameters
        ----------
        basis_matrix: np.ndarray, shape [n_features, n_samples]
            Matrix whose columns are the basis vectors in which to
            represent the measurement data.
        optimizer_kws: dictionary, optional
            Keyword arguments to be passed to the qr method.

        Returns
        -------
        self: a fitted :class:`pysensors.optimizers.CCQR` instance
        """

        n, m = basis_matrix.shape  # We transpose basis_matrix below

        if self.sensor_costs is None:
            self.sensor_costs = np.zeros(n)

        if len(self.sensor_costs) != n:
            raise ValueError(
                f"Dimension of sensor_costs ({len(self.sensor_costs)}) "
                f"does not match number of sensors in data ({n})"
            )

        # Initialize helper variables
        R = basis_matrix.conj().T.copy()
        p = np.arange(n)
        k = min(m, n)

        for j in range(k):
            u, i_piv = qr_reflector(R[j:, j:], self.sensor_costs[p[j:]])
            # Track column pivots
            i_piv += j
            p[[j, i_piv]] = p[[i_piv, j]]
            # Switch columns
            R[:, [j, i_piv]] = R[:, [i_piv, j]]
            # Apply reflector
            R[j:, j:] -= np.outer(u, np.dot(u, R[j:, j:]))
            R[j + 1 :, j] = 0

        self.pivots_ = p

        return self


def qr_reflector(r, costs):
    """
    Get the best (Householder) reflector with column pivoting and
    a cost function.

    The pivoting is biased by a cost function, i.e.
    the pivot is chosen as the argmax of :code:`norm(r[:, i]) - costs[i]`,
    whereas normally it would be chosen as the argmax of :code:`norm(r[:, i])`.

    Parameters
    ----------
    r: np.ndarray, shape [n_features, n_examples]
        Sub-array for which the pivot and reflector are to be found

    costs: np.ndarray, shape [n_examples,]
        Costs for each column (sensor location) in r

    Returns
    -------
    u: np.ndarray, shape [n_features,]
        Householder reflector.

    i_piv: nonnegative integer
        Index of the pivot.
    """

    # Norm of each column
    dlens = np.sqrt(np.sum(np.abs(r) ** 2, axis=0))

    # Choose pivot
    i_piv = np.argmax(dlens - costs)

    dlen = dlens[i_piv]

    if dlen > 0:
        u = r[:, i_piv] / dlen
        u[0] += np.sign(u[0]) + (u[0] == 0)
        u /= np.sqrt(abs(u[0]))
    else:
        u = r[:, i_piv]
        u[0] = np.sqrt(2)

    return u, i_piv



"""Tests for SSPOC class."""
import numpy as np
import pytest
from sklearn.datasets import make_classification
from sklearn.exceptions import NotFittedError
from sklearn.metrics import accuracy_score
from sklearn.utils.validation import check_is_fitted

from pysensors.basis import Identity
from pysensors.basis import RandomProjection
from pysensors.basis import SVD
from pysensors.classification import SSPOC


SEED = 15


@pytest.fixture
def data_binary_classification():
    x, y = make_classification(n_classes=2, random_state=SEED)
    l1_penalty = 1
    return x, y, l1_penalty


@pytest.fixture
def data_multiclass_classification():
    x, y = make_classification(n_classes=5, n_informative=5, random_state=SEED)
    l1_penalty = 0.03
    return x, y, l1_penalty


def test_not_fitted(data_binary_classification):
    x, y, _ = data_binary_classification
    model = SSPOC()

    # Shouldn't be able to call any of these methods before fitting
    with pytest.raises(NotFittedError):
        model.predict(x)
    with pytest.raises(NotFittedError):
        model.update_sensors(n_sensors=5)
    with pytest.raises(NotFittedError):
        model.selected_sensors


def test_prefit_basis(data_binary_classification):
    x, y, _ = data_binary_classification
    basis = Identity().fit(x)
    model_prefit = SSPOC(basis=basis)
    model_prefit.fit(x, y, prefit_basis=True, quiet=True)

    model_standard = SSPOC().fit(x, y, quiet=True)

    np.testing.assert_allclose(model_prefit.sensor_coef_, model_standard.sensor_coef_)


@pytest.mark.parametrize(
    "data",
    [
        pytest.lazy_fixture("data_binary_classification"),
        pytest.lazy_fixture("data_multiclass_classification"),
    ],
)
def test_initialize_with_n_sensors(data):
    x, y, l1_penalty = data
    n_sensors = 3
    model = SSPOC(n_sensors=n_sensors, l1_penalty=l1_penalty).fit(x, y, quiet=True)

    assert len(model.selected_sensors) == n_sensors
    assert model.n_sensors == n_sensors


@pytest.mark.parametrize(
    "data",
    [
        pytest.lazy_fixture("data_binary_classification"),
        pytest.lazy_fixture("data_multiclass_classification"),
    ],
)
def test_initialize_with_threshold(data):
    x, y, l1_penalty = data
    max_sensors = x.shape[1]
    model = SSPOC(threshold=0, l1_penalty=l1_penalty).fit(x, y, quiet=True)

    assert len(model.selected_sensors) == max_sensors
    assert model.n_sensors == max_sensors


@pytest.mark.parametrize("n_sensors", [3, 0])
@pytest.mark.parametrize(
    "data",
    [
        pytest.lazy_fixture("data_binary_classification"),
        pytest.lazy_fixture("data_multiclass_classification"),
    ],
)
def test_update_n_sensors(data, n_sensors):
    x, y, l1_penalty = data
    model = SSPOC(l1_penalty=l1_penalty).fit(x, y, quiet=True)

    model.update_sensors(n_sensors=n_sensors, quiet=True)
    assert len(model.selected_sensors) == n_sensors
    assert model.n_sensors == n_sensors


@pytest.mark.parametrize(
    "data",
    [
        pytest.lazy_fixture("data_binary_classification"),
        pytest.lazy_fixture("data_multiclass_classification"),
    ],
)
def test_update_threshold(data):
    x, y, l1_penalty = data

    model = SSPOC(threshold=0.01, l1_penalty=l1_penalty).fit(x, y, quiet=True)
    nnz = len(model.selected_sensors)

    # Larger threshold should result in fewer sensors
    model.update_sensors(threshold=1, quiet=True)
    assert len(model.selected_sensors) < nnz


@pytest.mark.parametrize(
    "data",
    [
        pytest.lazy_fixture("data_binary_classification"),
        pytest.lazy_fixture("data_multiclass_classification"),
    ],
)
def test_large_threshold(data):
    x, y, l1_penalty = data
    model = SSPOC(l1_penalty=l1_penalty).fit(x, y, quiet=True)

    model.update_sensors(threshold=10, quiet=True)
    assert len(model.selected_sensors) == 0
    assert model.n_sensors == 0


def test_bad_update_sensors_input(data_binary_classification):
    x, y, _ = data_binary_classification
    model = SSPOC().fit(x, y, quiet=True)

    with pytest.raises(ValueError):
        model.update_sensors()


@pytest.mark.parametrize(
    "data, baseline_accuracy",
    [
        (pytest.lazy_fixture("data_binary_classification"), 0.55),
        (pytest.lazy_fixture("data_multiclass_classification"), 0.25),
    ],
)
def test_predict_accuracy(data, baseline_accuracy):
    x, y, l1_penalty = data
    model = SSPOC(threshold=0, l1_penalty=l1_penalty).fit(x, y, quiet=True)

    assert (
        accuracy_score(y, model.predict(x[:, model.selected_sensors]))
        > baseline_accuracy
    )


@pytest.mark.parametrize(
    "data",
    [
        pytest.lazy_fixture("data_binary_classification"),
        pytest.lazy_fixture("data_multiclass_classification"),
    ],
)
def test_dummy_predict(data):
    x, y, l1_penalty = data

    model = SSPOC(l1_penalty=l1_penalty).fit(x, y, quiet=True)
    model.update_sensors(n_sensors=0, xy=(x, y), quiet=True)

    assert model.n_sensors == 0
    # Test that model can still make predictions, albeit random ones
    # when it has no sensors to work with
    y_pred = model.predict(x)
    assert len(y_pred) == len(y)


@pytest.mark.parametrize(
    "data",
    [
        pytest.lazy_fixture("data_binary_classification"),
        pytest.lazy_fixture("data_multiclass_classification"),
    ],
)
@pytest.mark.parametrize(
    "basis", [Identity(), SVD(), RandomProjection(n_basis_modes=5)]
)
def test_basis_integration(basis, data):
    x, y, _ = data
    model = SSPOC(basis=basis, n_sensors=5)
    model.fit(x, y, quiet=True)

    check_is_fitted(model)


@pytest.mark.parametrize(
    "data, shape",
    [
        (pytest.lazy_fixture("data_binary_classification"), (20,)),
        (pytest.lazy_fixture("data_multiclass_classification"), (20, 5)),
    ],
)
def test_coefficient_shape(data, shape):
    x, y, _ = data
    model = SSPOC().fit(x, y, quiet=True)

    assert model.sensor_coef_.shape == shape


@pytest.mark.parametrize("basis", [SVD, RandomProjection])
def test_update_n_basis_modes_errors(basis, data_binary_classification):
    x, y, _ = data_binary_classification
    n_basis_modes = 5
    model = SSPOC(basis=basis(n_basis_modes=n_basis_modes))

    model.fit(x, y, quiet=True)

    with pytest.raises(ValueError):
        model.update_n_basis_modes(0, xy=(x, y))
    with pytest.raises(ValueError):
        model.update_n_basis_modes("5", xy=(x, y))
    with pytest.raises(ValueError):
        model.update_n_basis_modes(x.shape[0] + 1, xy=(x, y))


@pytest.mark.parametrize("basis", [SVD, RandomProjection])
def test_update_n_basis_modes_shape(basis, data_binary_classification):
    x, y, _ = data_binary_classification
    n_basis_modes_init = 10
    model = SSPOC(basis=basis(n_basis_modes=n_basis_modes_init))
    model.fit(x, y, quiet=True)
    assert model.basis.n_basis_modes == n_basis_modes_init
    assert model.basis_matrix_inverse_.shape[0] == n_basis_modes_init

    n_basis_modes = 5
    model.update_n_basis_modes(n_basis_modes, xy=(x, y), quiet=True)
    assert model.basis.n_basis_modes == n_basis_modes_init
    assert model.basis_matrix_inverse_.shape[0] == n_basis_modes


@pytest.mark.parametrize("basis", [SVD, RandomProjection])
def test_update_n_basis_modes_refit(basis, data_binary_classification):
    x, y, _ = data_binary_classification
    n_basis_modes = 5
    model = SSPOC(basis=basis(n_basis_modes=n_basis_modes))
    model.fit(x, y, quiet=True)
    assert model.basis_matrix_inverse_.shape[0] == n_basis_modes

    model.update_n_basis_modes(n_basis_modes + 1, (x, y), quiet=True)
    assert model.basis_matrix_inverse_.shape[0] == n_basis_modes + 1


@pytest.mark.parametrize("basis", [SVD, RandomProjection])
def test_update_n_basis_modes_unfit_basis(basis, data_binary_classification):
    x, y, _ = data_binary_classification
    n_basis_modes = 5
    model = SSPOC(basis=basis())
    model.update_n_basis_modes(n_basis_modes, (x, y), quiet=True)

    assert model.basis_matrix_inverse_.shape[0] == n_basis_modes


def test_sspoc_selector_equivalence(data_multiclass_classification):
    x, y, _ = data_multiclass_classification

    model = SSPOC().fit(x, y)

    np.testing.assert_array_equal(model.get_selected_sensors(), model.selected_sensors)



"""
Sparse Sensor Placement Optimization for Classification (SSPOC) based
on

    Brunton, Bingni W., et al.
    "Sparse sensor placement optimization for classification."
    SIAM Journal on Applied Mathematics 76.5 (2016): 2099-2122.

See also the following paper for improvements on this method

    Mohren, Thomas L., et al.
    "Neural-inspired sensors enable sparse, efficient classification
    of spatiotemporal data."
    Proceedings of the National Academy of Sciences
    115.42 (2018): 10564-10569.
"""
import warnings

import numpy as np
from sklearn.base import BaseEstimator
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
from sklearn.dummy import DummyClassifier
from sklearn.exceptions import ConvergenceWarning
from sklearn.utils.validation import check_is_fitted

from ..basis import Identity
from ..utils import constrained_binary_solve
from ..utils import constrained_multiclass_solve
from ..utils import validate_input


INT_DTYPES = (int, np.int64, np.int32, np.int16, np.int8)


class SSPOC(BaseEstimator):
    r"""
    Sparse Sensor Placement Optimization for Classification (SSPOC) object.

    As the name suggests, this class can be used to select optimal sensor
    locations (measurement locations) for classification tasks.

    The time complexity of the SSPOC algorithm can be decomposed as

    .. math::

        C_{total} = C_{basis} + C_{classification} + C_{optimization}

    * :math:`C_{basis}`: the complexity of fitting the selected basis object
      and producing the matrix inverse. The matrix inverse is "free" to compute
      for :class:`pysensors.basis.Identity` and :class:`pysensors.basis.SVD`.
      For :class:`pysensors.basis.RandomProjection` the complexity is that
      of calling :code:`numpy.linalg.pinv` on a matrix of size
      :code:`n_input_features * n_basis_modes`.
    * :math:`C_{classification}`: the cost of fitting the chosen classifier
      to :code:`n_examples` examples with :code:`n_basis_modes` features.
    * :math:`C_{optimization}`: the cost of solving the sensor optimization
      problem. For binary classification we use
      :code:`sklearn.linear_model.OrthogonalMatchingPursuit`.
      For multi-class classification we use
      :code:`sklearn.linear_model.MultiTaskLasso`.
      The costs for each depend on the fit options that are specified.
      In both cases there are :code:`n_basis_modes` examples
      with :code:`n_features` features.

    The space complexity likewise depends on the same three factors.
    Generally, the basis requires :code:`O(n_basis_modes * n_features)`
    space. The space requirements for classification and optimization depend
    on the particular algorithms being employed. See the Scikit-learn
    documentation for specifics.

    See the following reference for more information:

        Brunton, Bingni W., et al.
        "Sparse sensor placement optimization for classification."
        SIAM Journal on Applied Mathematics 76.5 (2016): 2099-2122.

    Parameters
    ----------
    basis: basis object, optional (default :class:`pysensors.basis.Identity`)
        Basis in which to represent the data. Default is the identity basis
        (i.e. raw features).

    classifier: classifier object, optional \
            (default Linear Discriminant Analysis (LDA))
        Classifier for which to optimize sensors. Must be a *linear* classifier
        with a :code:`coef_` attribute and :code:`fit` and :code:`predict`
        methods.

    n_sensors: positive integer, optional (default None)
        Number of sensor locations to be used after fitting.
        If :code:`n_sensors` is not None then it overrides the :code:`threshold`
        parameter.
        If set to 0, then :code:`classifier` will be replaced with a dummy
        classifier which predicts the class randomly.

    threshold: nonnegative float, optional (default None)
        Threshold for selecting sensors.
        Overriden by :code:`n_sensors`.
        If both :code:`threshold` and :code:`n_sensors` are None when the
        :meth:`fit` method is called, then the threshold will be set to

        .. math::
            \frac{\|s\|_F}{2rc}

        where :math:`s` is a sensor coefficient matrix, :math:`r` is the number
        of basis modes, and :math:`c` is the number of distinct classes,
        as suggested in Brunton et al. (2016).

    l1_penalty: nonnegative float, optional (default 0.1)
        The L1 penalty term used to form the sensor coefficient matrix, s.
        Larger values will result in a sparser s and fewer selected sensors.
        This parameter is ignored for binary classification problems.

    Attributes
    ----------
    n_basis_modes: nonnegative integer
        Number of basis modes to be used when deciding sensor locations.

    basis_matrix_inverse_: np.ndarray, shape (n_basis_modes, n_input_features)
        The inverse of the matrix of basis vectors.

    sensor_coef_: np.ndarray, shape (n_input_features, n_classes)
        The sensor coefficient matrix, s.

    sparse_sensors_: np.ndarray, shape (n_sensors, )
        The selected sensors.

    Examples
    --------
    >>> from sklearn.metrics import accuracy_score
    >>> from sklearn.datasets import make_classification
    >>> from pysensors.classification import SSPOC
    >>>
    >>> x, y = make_classification(n_classes=3, n_informative=3, random_state=10)
    >>>
    >>> model = SSPOC(n_sensors=10, l1_penalty=0.03)
    >>> model.fit(x, y, quiet=True)
    SSPOC(basis=Identity(n_basis_modes=100),
          classifier=LinearDiscriminantAnalysis(), l1_penalty=0.03, n_sensors=10)
    >>> print(model.selected_sensors)
    [10 13  6 19 17 16 15 14 12 11]
    >>>
    >>> acc = accuracy_score(y, model.predict(x[:, model.selected_sensors]))
    >>> print("Accuracy:", acc)
    Accuracy: 0.66
    >>>
    >>> model.update_sensors(n_sensors=5, xy=(x, y), quiet=True)
    >>> print(model.selected_sensors)
    [10 13  6 19 17]
    >>>
    >>> acc = accuracy_score(y, model.predict(x[:, model.selected_sensors]))
    >>> print("Accuracy:", acc)
    Accuracy: 0.6
    """

    def __init__(
        self,
        basis=None,
        classifier=None,
        n_sensors=None,
        threshold=None,
        l1_penalty=0.1,
    ):
        if basis is None:
            basis = Identity()
        self.basis = basis
        if classifier is None:
            classifier = LinearDiscriminantAnalysis()
        self.classifier = classifier
        self.n_sensors = n_sensors
        self.threshold = threshold
        self.l1_penalty = l1_penalty
        self.n_basis_modes = None
        self.refit_ = False

    def fit(
        self,
        x,
        y,
        quiet=False,
        prefit_basis=False,
        refit=True,
        **optimizer_kws,
    ):
        """
        Fit the SSPOC model, determining which sensors are relevant.

        Parameters
        ----------
        x: array-like, shape (n_samples, n_input_features)
            Training data.

        y: array-like, shape (n_samples,)
            Training labels.

        quiet: boolean, optional (default False)
            Whether or not to suppress warnings during fitting.

        prefit_basis: boolean, optional (default False)
            Whether or not the basis has already been fit to x.
            For example, you may have already fit and experimented with
            a ``SVD`` object to determine the optimal number of modes. This
            option allows you to avoid an unnecessary SVD.

        refit: boolean, optional (default True)
            Whether or not to refit the classifier using measurements
            only from the learned sensor locations.

        optimizer_kws: dict, optional
            Keyword arguments to be passed to the optimization routine.

        Returns
        -------
        self: a fitted :class:`SSPOC` instance
        """

        # Fit basis functions to data
        if prefit_basis:
            check_is_fitted(self.basis, "basis_matrix_")
        else:
            x = validate_input(x)

            if quiet:
                with warnings.catch_warnings():
                    warnings.filterwarnings("ignore", category=UserWarning)
                    self.basis.fit(x)
            else:
                self.basis.fit(x)

        # Get matrix representation of basis - this is \Psi^T in the paper
        self.basis_matrix_inverse_ = self.basis.matrix_inverse(
            n_basis_modes=self.n_basis_modes
        )

        # Find weight vector
        if quiet:
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore", category=UserWarning)
                warnings.filterwarnings("ignore", category=ConvergenceWarning)
                self.classifier.fit(np.matmul(x, self.basis_matrix_inverse_.T), y)
        else:
            self.classifier.fit(np.matmul(x, self.basis_matrix_inverse_.T), y)

        w = np.squeeze(self.classifier.coef_).T

        n_classes = len(set(y[:]))
        if n_classes == 2:
            s = constrained_binary_solve(
                w, self.basis_matrix_inverse_, quiet=quiet, **optimizer_kws
            )
        else:
            s = constrained_multiclass_solve(
                w,
                self.basis_matrix_inverse_,
                alpha=self.l1_penalty,
                quiet=quiet,
                **optimizer_kws,
            )

        if self.threshold is None:
            # Chosen as in Brunton et al. (2016)
            threshold = np.sqrt(np.sum(s**2)) / (
                2 * self.basis_matrix_inverse_.shape[0] * n_classes
            )
        else:
            threshold = self.threshold

        # Decide which sensors to retain based on s
        self.sensor_coef_ = s
        self.sparse_sensors_ = np.array([])
        xy = (x, y) if refit else None
        self.update_sensors(
            n_sensors=self.n_sensors, threshold=threshold, xy=xy, quiet=quiet
        )

        # Form a dummy classifier for when no sensors are retained
        self.dummy_ = DummyClassifier(strategy="stratified")
        self.dummy_.fit(x[:, 0], y)

        return self

    def predict(self, x):
        """
        Predict classes for given measurements.
        If :code:`self.n_sensors` is 0 then a dummy classifier is used in place
        of :code:`self.classifier`.

        Parameters
        ----------
        x: array-like, shape (n_samples, n_sensors) or (n_samples, n_features)
            Examples to be classified.
            The measurements should be taken at the sensor locations specified by
            ``self.selected_sensors``.

        Returns
        -------
        y: np.ndarray, shape (n_samples,)
            Predicted classes.
        """
        check_is_fitted(self, "sensor_coef_")
        if self.n_sensors == 0:
            warnings.warn(
                "SSPOC model has no selected sensors so predictions are random. "
                "Increase n_sensors or lower threshold with SSPOC.update_sensors."
            )
            return self.dummy_.predict(x[:, 0])
        if self.refit_:
            return self.classifier.predict(x)
        else:
            return self.classifier.predict(np.dot(x, self.basis_matrix_inverse_.T))

    def update_sensors(
        self,
        n_sensors=None,
        threshold=None,
        xy=None,
        quiet=False,
        method=np.max,
        **method_kws,
    ):
        """
        Update the selected sensors by changing either the preferred number of sensors
        or the threshold used to select the sensors, refitting the classifier
        afterwards, if possible.

        Parameters
        ----------
        n_sensors: nonnegative integer, optional (default None)
            The number of sensor locations to select.
            If None, then :code:`threshold` will be used to pick the sensors.
            Note that :code:`n_sensors` and :code:`threshold` cannot both be None.

        threshold: nonnegative float, optional (default None)
            The threshold to use to select sensors based on the magnitudes of entries
            in :code:`self.sensor_coef_` (s).
            Overridden by :code:`n_sensors`.
            Note that :code:`n_sensors` and :code:`threshold` cannot both be None.

        xy: tuple of np.ndarray, length 2, optional (default None)
            Tuple containing training data x and labels y for refitting.
            x should have shape (n_samples, n_input_features) and y shape (n_samples, ).
            If not None, the classifier will be refit after the new sensors have been
            selected.

        quiet: boolean, optional (default False)
            Whether to silence warnings.

        method: callable, optional (default :code:`np.max`)
            Function used along with :code:`threshold` to select sensors.
            For binary classification problems one need not specify a method.
            For multiclass classification problems, :code:`sensor_coef_` (s) has
            multiple columns and :code:`method` is applied along each row to aggregate
            coefficients for thresholding, i.e. :code:`method` is called as follows
            :code:`method(np.abs(self.sensor_coef_), axis=1, **method_kws)`.
            Other examples of acceptable methods are :code:`np.min`, :code:`np.mean`,
            and :code:`np.median`.

        **method_kws: dict, optional
            Keyword arguments to be passed into :code:`method` when it is called.
        """
        check_is_fitted(self, "sensor_coef_")
        warn = not quiet

        if n_sensors is not None and threshold is not None and warn:
            warnings.warn(
                f"Both n_sensors({n_sensors}) and threshold({threshold}) "
                "were passed so threshold will be ignored"
            )

        if n_sensors is None and threshold is None:
            raise ValueError("At least one of n_sensors or threshold must be passed.")

        elif n_sensors is not None:
            if n_sensors > len(self.sensor_coef_):
                raise ValueError(
                    f"n_sensors({n_sensors}) cannot exceed number of available "
                    f"sensors ({len(self.sensor_coef_)})"
                )
            self.n_sensors = n_sensors
            # Could be made more efficient with a max heap
            # (we don't need to sort the whole list)
            if np.ndim(self.sensor_coef_) == 1:
                sorted_sensors = np.argsort(-np.abs(self.sensor_coef_))
                if (
                    np.abs(self.sensor_coef_[sorted_sensors[n_sensors - 1]]) == 0
                    and warn
                ):
                    warnings.warn(
                        "Some uninformative sensors were selected. "
                        "Consider decreasing n_sensors"
                    )
            else:
                sorted_sensors = np.argsort(
                    -method(np.abs(self.sensor_coef_), axis=1, **method_kws)
                )
                if (
                    method(
                        np.abs(self.sensor_coef_[sorted_sensors[n_sensors - 1], :]),
                        **method_kws,
                    )
                    == 0
                    and warn
                ):
                    warnings.warn(
                        "Some uninformative sensors were selected. "
                        "Consider decreasing n_sensors"
                    )
            self.sparse_sensors_ = sorted_sensors[:n_sensors]

        else:
            self.threshold = threshold
            if np.ndim(self.sensor_coef_) == 1:
                sparse_sensors = np.nonzero(np.abs(self.sensor_coef_) >= threshold)[0]
            else:
                sparse_sensors = np.nonzero(
                    method(np.abs(self.sensor_coef_), axis=1, **method_kws) >= threshold
                )[0]

            self.n_sensors = len(sparse_sensors)
            self.sparse_sensors_ = sparse_sensors

            if self.n_sensors == 0 and warn:
                warnings.warn(
                    f"Threshold set too high ({threshold}); no sensors selected."
                )

        # Refit if xy was passed
        if xy is not None:
            if self.n_sensors > 0:
                x, y = xy

                if quiet:
                    # Suppress warnings arising from no sensors being selected
                    with warnings.catch_warnings():
                        warnings.filterwarnings("ignore", category=UserWarning)
                        warnings.filterwarnings("ignore", category=ConvergenceWarning)
                        self.classifier.fit(x[:, self.sparse_sensors_], y)
                else:
                    self.classifier.fit(x[:, self.sparse_sensors_], y)
                self.refit_ = True
            else:
                warnings.warn("No selected sensors; model was not refit.")

    def update_n_basis_modes(self, n_basis_modes, xy, **fit_kws):
        """
        Re-fit the :class:`SSPOC` object using a different value of
        :code:`n_basis_modes`.

        This method allows one to relearn sensor locations for a
        different number of basis modes _without_ re-fitting the basis
        in many cases.
        Specifically, if :code:`n_basis_modes <= self.basis.n_basis_modes`
        then the basis does not need to be refit.
        Otherwise this function does not save any computational resources.

        Parameters
        ----------
        n_basis_modes: positive int, optional (default None)
            Number of basis modes to be used during fit.
            Must be less than or equal to ``n_samples``.

        xy: tuple of np.ndarray, length 2
            Tuple containing training data x and labels y for refitting.
            x should have shape (n_samples, n_input_features) and y shape (n_samples, ).

        **fit_kws: dict, optional
            Keyword arguments to pass to :meth:`SSPOC.fit`.
        """
        if not isinstance(n_basis_modes, INT_DTYPES) or n_basis_modes <= 0:
            raise ValueError("n_basis_modes must be a positive integer")

        x, y = xy
        # No need to refit basis; only refit sensors
        if (
            hasattr(self.basis, "basis_matrix_")
            and n_basis_modes <= self.basis.n_basis_modes
        ):
            self.n_basis_modes = n_basis_modes
            self.fit(x, y, prefit_basis=True, **fit_kws)

        else:
            if n_basis_modes > x.shape[0]:
                raise ValueError(
                    "n_basis_modes cannot exceed the number of examples, x.shape[0]"
                )
            else:
                self.n_basis_modes = n_basis_modes
                self.basis.n_basis_modes = n_basis_modes
                self.fit(x, y, prefit_basis=False, **fit_kws)

    @property
    def selected_sensors(self):
        """
        Get the indices of the selected sensors.

        Returns
        -------
        sensors: numpy array, shape (n_sensors,)
            Indices of the selected sensors.
        """
        check_is_fitted(self, "sparse_sensors_")
        return self.sparse_sensors_

    def get_selected_sensors(self):
        """
        Convenience function for getting indices of the selected sensors.

        Returns
        -------
        sensors: numpy array, shape (n_sensors,)
            Indices of the selected sensors.
        """
        return self.selected_sensors



"""Unit tests for basis classes"""
import warnings

import numpy as np
import pytest
from sklearn.exceptions import NotFittedError

from pysensors.basis import Identity
from pysensors.basis import RandomProjection
from pysensors.basis import SVD


@pytest.mark.parametrize("basis", [Identity(), SVD(), RandomProjection()])
def test_not_fitted(basis):
    with pytest.raises(NotFittedError):
        basis.matrix_representation()


def test_identity_matrix_representation(data_random):
    matrix = data_random

    basis = Identity()
    basis.fit(matrix)

    np.testing.assert_allclose(matrix.T, basis.matrix_representation())


@pytest.mark.parametrize("basis", [SVD, RandomProjection])
def test_matrix_representation(basis, data_random):
    data = data_random
    n_features = data.shape[1]
    n_components = 5

    b = basis(n_basis_modes=n_components)
    b.fit(data)
    matrix_representation = b.matrix_representation()

    assert matrix_representation.shape[0] == n_features
    assert matrix_representation.shape[1] == n_components


def test_random_projection_random_state(data_vandermonde):
    data = data_vandermonde

    basis1 = RandomProjection(n_basis_modes=5, random_state=1)
    m1 = basis1.fit(data).matrix_representation()

    basis2 = RandomProjection(n_basis_modes=5, random_state=2)
    m2 = basis2.fit(data).matrix_representation()

    assert not np.allclose(m1, m2)


@pytest.mark.parametrize("basis", [Identity, SVD, RandomProjection])
def test_n_basis_modes(basis, data_random):
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=UserWarning)

        with pytest.raises(ValueError):
            b = basis(n_basis_modes=0)
        with pytest.raises(ValueError):
            b = basis(n_basis_modes=1.2)
        with pytest.raises(ValueError):
            b = basis(n_basis_modes="1")

        data = data_random
        n_basis_modes = 5
        b = basis(n_basis_modes=n_basis_modes)
        b.fit(data)

        assert b.matrix_representation().shape[1] == n_basis_modes


@pytest.mark.parametrize("basis", [Identity, SVD])
def test_extra_basis_modes(basis, data_random):
    data = data_random
    n_basis_modes = data.shape[0] + 1
    b = basis(n_basis_modes=n_basis_modes)
    # Can't have more basis modes than the number of training examples
    with pytest.raises(ValueError):
        b.fit(data)


@pytest.mark.parametrize("basis", [SVD(), RandomProjection()])
def test_matrix_inverse_shape(basis, data_random):
    data = data_random
    n_features = data.shape[1]
    n_basis_modes = 5

    basis.fit(data)
    inverse = basis.matrix_inverse(n_basis_modes=n_basis_modes)

    assert inverse.shape == (n_basis_modes, n_features)



import warnings

from numpy import ndim
from sklearn.linear_model import MultiTaskLasso
from sklearn.linear_model import OrthogonalMatchingPursuit


def constrained_binary_solve(
    w, psi, quiet=False, fit_intercept=True, normalize=True, precompute="auto"
):
    if ndim(w) != 1:
        raise ValueError(
            f"w must be a 1D vector; received a vector of dimension {ndim(w)}"
        )

    model = OrthogonalMatchingPursuit(
        tol=0, fit_intercept=fit_intercept, normalize=normalize, precompute=precompute
    )

    if quiet:
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=RuntimeWarning)
            warnings.filterwarnings("ignore", category=UserWarning)
            model.fit(psi, w)
    else:
        model.fit(psi, w)

    return model.coef_


def constrained_multiclass_solve(w, psi, alpha=1.0, quiet=False, **lasso_kws):
    """
    Solve

    .. math::

        \\text{argmin}_s \\|s\\|_0 \\\\
        \\text{subject to} \\|w - \\psi s\\|_2^2 \\leq tol
    """
    model = MultiTaskLasso(alpha=alpha, **lasso_kws)

    if quiet:
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=RuntimeWarning)
            warnings.filterwarnings("ignore", category=UserWarning)
            model.fit(psi, w)
    else:
        model.fit(psi, w)

    return model.coef_.T



"""
SVD mode basis class.
"""
from sklearn.decomposition import TruncatedSVD

from ._base import InvertibleBasis
from ._base import MatrixMixin


class SVD(TruncatedSVD, InvertibleBasis, MatrixMixin):
    """
    Generate an SVD transformation which maps input features to
    SVD modes.

    Assumes the data has already been centered (to have mean 0).

    Parameters
    ----------
    n_basis_modes : int, optional (default 10)
        Number of basis modes to retain. Cannot be larger than
        the number of features ``n_features``, or the number of examples
        ``n_examples``.

    algorithm: string, optional (default "randomized")
        SVD solver to use. Either “arpack” for the ARPACK wrapper
        in SciPy (scipy.sparse.linalg.svds), or “randomized” for the
        randomized algorithm due to Halko (2009).

    kwargs : dict, optional
        Keyword arguments to be passed to the TruncatedSVD constructor

    Attributes
    ----------
    basis_matrix_ : numpy ndarray, shape (n_features, n_basis_modes)
        The top n_basis_modes left singular vectors of the training data.

    References
    -----
    https://scikit-learn.org/stable/modules/generated/\
    sklearn.decomposition.TruncatedSVD.html

    Finding structure with randomness: Stochastic algorithms for
    constructing approximate matrix decompositions Halko, et al.,
    2009 (arXiv:909) https://arxiv.org/pdf/0909.4061.pdf
    """

    def __init__(self, n_basis_modes=10, **kwargs):
        if isinstance(n_basis_modes, int) and n_basis_modes > 0:
            super(SVD, self).__init__(n_components=n_basis_modes, **kwargs)
            self._n_basis_modes = n_basis_modes
        else:
            raise ValueError("n_basis_modes must be a positive integer.")

    def fit(self, X):
        """
        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            The training data.

        Returns
        -------
        self : instance
        """
        self.basis_matrix_ = super(SVD, self).fit(X).components_.T
        return self

    def matrix_inverse(self, n_basis_modes=None):
        """
        Get the inverse matrix mapping from measurement space to
        coordinates with respect to the basis.

        Note that this is not the inverse of the matrix returned by
        ``self.matrix_representation``. It is the (psuedo) inverse of
        the matrix whose columns are the basis modes.

        Parameters
        ----------
        n_basis_modes : positive int, optional (default None)
            Number of basis modes to be used to compute inverse.

        Returns
        -------
        B : numpy ndarray, shape (n_basis_modes, n_features)
            The inverse matrix.
        """
        n_basis_modes = self._validate_input(n_basis_modes)

        return self.basis_matrix_[:, :n_basis_modes].T

    @property
    def n_basis_modes(self):
        """Number of basis modes."""
        return self._n_basis_modes

    @n_basis_modes.setter
    def n_basis_modes(self, n_basis_modes):
        self._n_basis_modes = n_basis_modes
        self.n_components = n_basis_modes



"""
Random projections basis class.

Project data onto random features.
"""
from numpy.linalg import pinv
from sklearn.random_projection import GaussianRandomProjection

from ._base import InvertibleBasis
from ._base import MatrixMixin


class RandomProjection(GaussianRandomProjection, InvertibleBasis, MatrixMixin):
    """
    Generate a basis based on Gaussian random projection.

    Wrapper for sklearn.random_projection.GaussianRandomProjection.

    Parameters
    ----------
    n_basis_modes : int or "auto", optional (default 10)
        Dimensionality of the target projection space (number of modes).

        `n_basis_modes` can be automatically adjusted according to the
        number of samples in the dataset and the bound given by the
        Johnson-Lindenstrauss lemma. In that case the quality of the
        embedding is controlled by the eps parameter.

    eps: positive float, optional (default 0.1)
        Parameter to control the quality of the embedding according to
        the Johnson-Lindenstrauss lemma when n_basis_modes is set to ‘auto’.

        Smaller values lead to better embedding and higher number of
        dimensions (n_basis_modes) in the target projection space.

    random_state: int, RandomState instance or None, optional (default=None)
        Controls the pseudo random number generator used to generate the
        projection matrix at fit time. Pass an int for reproducible output
        across multiple function calls.

    Attributes
    ----------
    basis_matrix_ : numpy ndarray, shape (n_features, n_basis_modes)
        Random projections of training data.

    References
    -----
    https://scikit-learn.org/stable/modules/generated/\
    sklearn.random_projection.GaussianRandomProjection.html

    """

    def __init__(self, n_basis_modes=10, eps=0.1, random_state=None):
        if (
            isinstance(n_basis_modes, int) and n_basis_modes > 0
        ) or n_basis_modes == "auto":
            super(RandomProjection, self).__init__(
                n_components=n_basis_modes, eps=eps, random_state=random_state
            )
            self.n_basis_modes = n_basis_modes
        else:
            raise ValueError("n_basis_modes must be a positive int or 'auto'")

    def fit(self, X):
        """
        X : array-like, shape (n_samples, n_features)
            The training data.

        Returns
        -------
        self : instance
        """

        super(RandomProjection, self).fit(X.T)
        self.basis_matrix_ = super(RandomProjection, self).transform(X.T)
        return self

    def matrix_inverse(self, n_basis_modes=None, **kwargs):
        """
        Get the inverse matrix mapping from measurement space to
        coordinates with respect to the basis.

        Note that this is not the inverse of the matrix returned by
        ``self.matrix_representation``. It is the (psuedo) inverse of
        the matrix whose columns are the basis modes.

        Parameters
        ----------
        n_basis_modes : positive int, optional (default None)
            Number of basis modes to be used to compute inverse.

        Returns
        -------
        B : numpy ndarray, shape (n_basis_modes, n_features)
            The inverse matrix.
        """
        n_basis_modes = self._validate_input(n_basis_modes)

        return pinv(self.basis_matrix_[:, :n_basis_modes], **kwargs)

    @property
    def n_basis_modes(self):
        """Number of basis modes."""
        return self._n_basis_modes

    @n_basis_modes.setter
    def n_basis_modes(self, n_basis_modes):
        self._n_basis_modes = n_basis_modes
        self.n_components = n_basis_modes
